from functools import wraps, partial, singledispatch
from inspect import getfullargspec
import typing
from typing import Callable, Union, Any, Dict

import fn

from amino.util.string import snake_case


class F(fn.F):

    def __truediv__(self, f):
        return self >> (lambda a: a / f)

    def __floordiv__(self, f):
        return self >> (lambda a: a // f)

    @property
    def name(self):
        from amino.anon import AnonCallable
        f = (
            self.f.func
            if isinstance(self.f.func, AnonCallable) else
            self.f.func.__name__
        ) if self._is_partial else self.f.__name__
        return str(f)

    @property
    def _is_partial(self):
        return isinstance(self.f, partial)

    def __str__(self):
        from amino.anon import format_funcall
        rep = (format_funcall(self.name, self.f.args, self.f.keywords)
               if self._is_partial
               else '{}()'.format(self.name))
        return 'F({})'.format(rep)


def curried(func):
    @wraps(func)
    def _curried(*args, **kwargs):
        f = func
        count = 0
        while isinstance(f, partial):
            if f.args:
                count += len(f.args)
            f = f.func
        spec = getfullargspec(f)
        if count == len(spec.args) - len(args):
            return func(*args, **kwargs)
        else:
            return curried(partial(func, *args, **kwargs))
    return _curried


class Identity:

    def __call__(self, a):
        return a

    def __str__(self):
        return '(a => a)'

I = Identity()


class Val:

    def __init__(self, value) -> None:
        self.value = value

    def __call__(self):
        return self.value

    def __str__(self):
        return '{}({})'.format(self.__class__.__name__, self.value)


class ReplaceVal(Val):

    def __call__(self, *a, **kw):
        return super().__call__()


def flip(a, b):
    return b, a

CallByName = Union[Any, Callable[[], Any]]


def call_by_name(b: CallByName):
    return b() if callable(b) else b


def dispatch(obj: Any, tpes: typing.List[type], prefix: str,
             default: Callable=None) -> Any:
    @singledispatch
    def main(o, *a, **kw):
        if default is None:
            msg = 'no dispatcher defined for {} on {} {}'
            raise TypeError(msg.format(o, obj.__class__.__name__, prefix))
        else:
            default(o, *a, **kw)
    for tpe in tpes:
        fun = getattr(obj, '{}{}'.format(prefix, snake_case(tpe.__name__)))
        main.register(tpe)(fun)
    return main


def dispatch_with(rules: Dict[type, Callable], default: Callable=None):
    @singledispatch
    def main(o, *a, **kw):
        if default is None:
            msg = 'no dispatcher defined for {} {} ({})'
            raise TypeError(msg.format(type(o), o, rules))
        else:
            default(o, *a, **kw)
    for tpe, fun in rules.items():
        main.register(tpe)(fun)
    return main

def is_not_none(a):
    return a is not None

__all__ = ('curried', 'F', 'I', 'flip', 'call_by_name', 'Val', 'ReplaceVal',
           'dispatch', 'dispatch_with', 'is_not_none')



import itertools
import typing
import random
import string
from functools import reduce
from typing import TypeVar, Callable, Generic, Iterable, Any

from toolz.itertoolz import cons, groupby

from amino import maybe, boolean
from amino.logging import log
from amino.tc.base import ImplicitsMeta, Implicits
from amino.func import curried, I
from amino.util.string import safe_string

A = TypeVar('A', covariant=True)
B = TypeVar('B')


def flatten(l: Iterable[Iterable[A]]) -> Iterable[A]:
    return list(itertools.chain.from_iterable(l))  # type: ignore


class ListMeta(ImplicitsMeta):

    def __instancecheck__(self, instance):
        if type(instance) is list:
            return False
        else:
            return super().__instancecheck__(instance)

    # def __subclasscheck__(self, subclass):
    #     if subclass is list:
    #         return False
    #     return super().__subclasscheck__(subclass)


def _rand_str(chars, num: int=10):
    return ''.join(random.choice(chars) for i in range(num))


class List(typing.List[A], Generic[A], Implicits, implicits=True,
           metaclass=ListMeta):

    def __init__(self, *elements):
        typing.List.__init__(self, elements)

    def __getitem__(self, arg):
        s = super().__getitem__(arg)
        return List.wrap(s) if isinstance(arg, slice) else s

    @staticmethod
    def wrap(l: Iterable[B]) -> 'List[B]':
        return List(*list(l))

    @staticmethod
    def range(*a):
        return List.wrap(range(*a))

    @staticmethod
    def random_string(num: int=10):
        chars = string.ascii_letters + string.digits
        return _rand_str(chars, num)

    @staticmethod
    def random_alpha(num: int=10):
        chars = string.ascii_letters
        return _rand_str(chars, num)

    @staticmethod
    def gen(num: int, f: Callable[[], A]):
        return List.range(num) // (lambda a: f())

    @staticmethod
    def lines(data: str):
        return List.wrap(data.splitlines())

    def lift(self, index: int) -> 'maybe.Maybe[A]':
        return (
            (maybe.Just(self[index]) if len(self) > index else maybe.Empty())
            if index >= 0 else
            (maybe.Just(self[index]) if len(self) >= -index else maybe.Empty())
        )

    def lift_all(self, index, *indexes):
        from amino.anon import _
        def folder(z, n):
            return n.ap(z / _.cat)
        els = List.wrap(indexes) / self.lift
        init = self.lift(index) / List
        return els.fold_left(init)(folder)

    def smap(self, f: Callable[..., B]) -> 'List[B]':
        return List.wrap(list(itertools.starmap(f, self)))

    def flat_smap(self, f: Callable[..., 'Iterable[B]']) -> 'List[B]':
        return List.wrap(flatten(list(itertools.starmap(f, self))))

    def foreach(self, f: Callable[[A], B]) -> None:
        for el in self:
            f(el)

    def forall(self, f: Callable[[A], bool]) -> boolean.Boolean:
        return boolean.Boolean(all(f(el) for el in self))

    def contains(self, value):
        return value in self

    def exists(self, f: Callable[[A], bool]) -> bool:
        return self.find(f).is_just

    @property
    def is_empty(self):
        return boolean.Boolean(self.length == 0)

    empty = is_empty

    @property
    def nonempty(self) -> boolean.Boolean:
        return not self.empty

    @property
    def length(self):
        return len(self)

    @property
    def head(self):
        return self.lift(0)

    @property
    def last(self):
        return self.lift(-1)

    @property
    def init(self):
        return maybe.Empty() if self.empty else maybe.Just(self[:-1])

    @property
    def tail(self):
        return maybe.Empty() if self.empty else maybe.Just(self[1:])

    @property
    def detach_head(self):
        return self.head.product(self.tail)

    @property
    def detach_last(self):
        return self.last.product(self.init)

    @property
    def distinct(self) -> 'List[A]':
        return self.distinct_by(I)

    def distinct_by(self, f: Callable[[A], bool]) -> 'List[A]':
        seen = set()  # type: set
        def pred(a):
            v = f(a)
            return v in seen or seen.add(v)
        return List.wrap(x for x in self if not pred(x))

    def add(self, other: typing.List[A]) -> 'List[A]':
        return List.wrap(typing.List.__add__(self, other))

    __add__ = add

    def without(self, el) -> 'List[A]':
        from amino.anon import _
        return self.filter(_ != el)

    __sub__ = without

    def split(self, f: Callable[[A], bool]):
        def splitter(z, e):
            l, r = z
            return (l + (e,), r) if f(e) else (l, r + (e,))
        l, r = reduce(splitter, self, ((), (),))
        return List.wrap(l), List.wrap(r)

    def split_type(self, tpe: type):
        return self.split(lambda a: isinstance(a, tpe))

    def debug(self, prefix=None):
        prefix = '' if prefix is None else prefix + ' '
        log.debug(prefix + str(self))
        return self

    def index_of(self, target: Any):
        from amino.anon import _
        return self.index_where(_ == target)

    @property
    def reversed(self):
        return List.wrap(reversed(self))

    def mk_string(self, sep: str='') -> str:
        return sep.join(self / str)

    @property
    def join_lines(self) -> str:
        return self.mk_string('\n')

    @property
    def join_comma(self) -> str:
        return self.mk_string(', ')

    @property
    def join_tokens(self) -> str:
        return self.mk_string(' ')

    @property
    def join_dot(self) -> str:
        return self.mk_string('.')

    def cons(self, item):
        return List.wrap(cons(item, self))

    def cat(self, item):
        return self + List(item)

    def cat_m(self, item: maybe.Maybe):
        return item / self.cat | self

    @property
    def transpose(self):
        return List.wrap(map(List.wrap, zip(*self)))

    def drop(self, n: int) -> 'List[A]':
        return self[n:]

    def take(self, n: int):
        return self[:n]

    def drop_while(self, pred: Callable[[A], bool]):
        index = self.index_where(lambda a: not pred(a))
        return index / (lambda a: self[a:]) | self

    def drop_right(self, n: int) -> 'List[A]':
        return self.take(self.length - n)

    def remove_all(self, els: 'List[A]') -> 'List[A]':
        return self.filter_not(els.contains)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               ', '.join(map(repr, self)))

    def __str__(self):
        return '[{}]'.format(', '.join(map(safe_string, self)))

    def __hash__(self):
        return hash(tuple(self))

    def sort_by(self, f: Callable[[A], bool], reverse=False):
        return List.wrap(sorted(self, key=f, reverse=reverse))

    def sort(self, reverse=False):
        return self.sort_by(I, reverse)

    def replace_item(self, a, b) -> 'List[A]':
        return self.map(lambda c: b if c == a else c)

    @curried
    def replace_where(self, a, pred: Callable):
        return self.map(lambda b: a if pred(b) else b)

    def __mul__(self, other):
        return List.wrap(super().__mul__(other))

    def group_by(self, f):
        from amino import Map
        return Map(groupby(f, self)).valmap(List.wrap)

    def slice(self, start: int, end: int) -> 'List[A]':
        return self[start:end]


class Lists:

    wrap = List.wrap
    range = List.range
    random_string = List.random_string
    random_alpha = List.random_alpha
    gen = List.gen
    lines = List.lines

    @staticmethod
    def split(data: str, splitter: str, maxsplit: int=-1) -> List[str]:
        return List.wrap(data.split(splitter, maxsplit))

    @staticmethod
    def rsplit(data: str, splitter: str, maxsplit: int=-1) -> List[str]:
        return List.wrap(data.rsplit(splitter, maxsplit))

__all__ = ('List',)



import os
import re
import logging
import operator
from typing import Callable, Any
import sys
from pathlib import Path

from amino.lazy import lazy

import amino
from amino.func import call_by_name

VERBOSE = 15
DDEBUG = 5
logging.addLevelName(VERBOSE, 'VERBOSE')
logging.addLevelName(DDEBUG, 'DDEBUG')


class Logger(logging.Logger):

    def verbose(self, message, *args, **kws):
        if self.isEnabledFor(VERBOSE):
            self._log(VERBOSE, message, args, **kws)

    def ddebug(self, message, *args, **kws) -> None:
        if self.isEnabledFor(DDEBUG):
            self._log(DDEBUG, call_by_name(message), args, **kws)

    def caught_exception(self, when, exc, *a, **kw):
        headline = 'caught exception while {}:'.format(when)
        self.exception(headline, exc_info=(type(exc), exc, exc.__traceback__))


logging.Logger.verbose = Logger.verbose  # type: ignore
logging.Logger.ddebug = Logger.ddebug  # type: ignore
logging.Logger.caught_exception = Logger.caught_exception  # type: ignore

log = amino_root_logger = logging.getLogger('amino')
log.setLevel(DDEBUG)


def install_logger_class():
    logging.setLoggerClass(Logger)


def amino_logger(name: str):
    return amino_root_logger.getChild(name)

_stdout_logging_initialized = False

_level_env_var = 'AMINO_LOG_LEVEL'


def init_loglevel(handler: logging.Handler, level: int=None):
    if level is not None:
        handler.setLevel(level)
    elif _level_env_var in os.environ:
        handler.setLevel(os.environ[_level_env_var])
    elif amino.development:
        handler.setLevel(VERBOSE)

amino_stdout_handler = logging.StreamHandler(stream=sys.stdout)


def amino_stdout_logging(level: int=None):
    global _stdout_logging_initialized
    if not _stdout_logging_initialized:
        amino_root_logger.addHandler(amino_stdout_handler)
        init_loglevel(amino_stdout_handler, level)
        _stdout_logging_initialized = True


default_logfile = Path.home() / '.python' / 'log'
_file_fmt = ('{asctime} [{levelname} @ {name}:{funcName}:{lineno}] {message}')


def amino_file_logging(logger: logging.Logger, level: int=logging.DEBUG,
                       logfile: Path=default_logfile, fmt: str=None) -> None:
    logfile.parent.mkdir(exist_ok=True)
    formatter = logging.Formatter(fmt or _file_fmt, style='{')
    handler = logging.FileHandler(str(logfile))
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    init_loglevel(handler, level)


def amino_root_file_logging(level: int=logging.DEBUG, **kw: Any) -> None:
    amino_file_logging(amino_root_logger, level, **kw)


class Logging:

    @property
    def log(self) -> Logger:
        return self._log

    @lazy
    def _log(self) -> Logger:
        return amino_logger(self.__class__.__name__)

    def _p(self, a):
        v = self.log.verbose
        v(a)
        return a

    def _dbg(self, fmt, level=DDEBUG):
        def log(a):
            msg = fmt.format(a)
            self.log.log(level, msg)
            return a
        return log


def sub_loggers(loggers, root):
    from amino import Map, _, L
    children = loggers.keyfilter(L(re.match)('{}\.[^.]+$'.format(root), _))
    sub = (children.k / L(sub_loggers)(loggers, _))\
        .fold_left(Map())(operator.pow)
    return Map({loggers[root]: sub})


def logger_tree(root):
    from amino import __, Map
    m = Map(logging.Logger.manager.loggerDict)
    all = m.keyfilter(__.startswith(root))
    return sub_loggers(all, 'amino')


def indent(strings, level, width=1):
    ws = ' ' * level * width
    return strings.map(ws.__add__)


def format_logger_tree(tree, fmt_logger, level=0):
    from amino import _, L
    sub_f = L(format_logger_tree)(_, fmt_logger, level=level + 1)
    formatted = tree.bimap(fmt_logger, sub_f)
    return '\n'.join(indent(formatted.map2('{}\n{}'.format), level))


def print_info(out: Callable[[str], None]):
    lname = lambda l: logging.getLevelName(l.getEffectiveLevel())
    hlname = lambda h: logging.getLevelName(h.level)
    def handler(h):
        return '{}({})'.format(h.__class__.__name__, hlname(h))
    def logger(l):
        handlers = ','.join(list(map(handler, l.handlers)))
        return '{}: {} {}'.format(l.name, lname(l), handlers)
    out(format_logger_tree(logger_tree('amino'), logger))

__all__ = ('amino_root_logger', 'amino_stdout_logging', 'amino_file_logging',
           'amino_root_file_logging')



import re
from functools import singledispatch  # type: ignore


def snake_case(name: str) -> str:
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


@singledispatch
def decode(value):
    return value


@decode.register(bytes)
def decode_bytes(value):
    return value.decode()


@decode.register(list)
def decode_list(value):
    from amino import List
    return List.wrap(value).map(decode)


@decode.register(dict)
def decode_dict(value):
    from amino import Map
    return Map.wrap(value)\
        .keymap(decode)\
        .valmap(decode)


@decode.register(Exception)
def decode_exc(value):
    return decode_list(value.args).head | str(value)


def camelcaseify(name, sep='', splitter='_'):
    return sep.join([n.capitalize() for n in re.split(splitter, name)])


def safe_string(value):
    try:
        return str(value)
    except Exception:
        try:
            return repr(value)
        except Exception:
            return 'invalid'

__all__ = ('snake_case', 'decode', 'camelcaseify')



from .spec import Spec, IntegrationSpec, later  # NOQA
from .path import (setup, create_temp_file, temp_file, temp_path, temp_dir,
                   fixture_path, load_fixture)  # NOPQ

__all__ = ('Spec', 'IntegrationSpec', 'create_temp_file', 'temp_file',
           'temp_path', 'temp_dir', 'fixture_path', 'load_fixture', 'setup',
           'later')



import importlib
from typing import TypeVar, Generic, Callable, Union, Any
from typing import Tuple  # NOQA

from amino import boolean
from amino.func import I
from amino.tc.base import Implicits

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')


class Either(Generic[A, B], Implicits, implicits=True):

    def __init__(self, value: Union[A, B]) -> None:
        self.value = value

    @staticmethod
    def import_name(modname, name):
        try:
            mod = importlib.import_module(modname)
        except ImportError as e:
            return Left(e)
        else:
            if hasattr(mod, name):
                return Right(getattr(mod, name))
            else:
                return Left('{} has no attribute {}'.format(mod, name))

    @staticmethod
    def import_path(path):
        from amino.list import List
        return (
            List.wrap(path.rsplit('.', 1))
            .lift_all(0, 1)
            .to_either('invalid module path: {}'.format(path))
            .flat_map2(Either.import_name)
        )

    @staticmethod
    def import_module(modname):
        try:
            mod = importlib.import_module(modname)
        except ImportError as e:
            return Left(e)
        else:
            return Right(mod)

    @property
    def is_right(self):
        return boolean.Boolean(isinstance(self, Right))

    @property
    def is_left(self):
        return boolean.Boolean(isinstance(self, Left))

    def leffect(self, f):
        if self.is_left:
            f(self.value)
        return self

    def bieffect(self, l: Callable[[A], Any],
                 r: Callable[[B], Any]) -> 'Either[A, B]':
        self.cata(l, r)
        return self

    def cata(self, fl: Callable[[A], Any], fr: Callable[[B], Any]):
        f = fl if self.is_left else fr
        return f(self.value)  # type: ignore

    def recover_with(self, f: Callable[[A], 'Either[A, B]']):
        return self.cata(f, Right)

    def right_or_map(self, f: Callable[[A], Any]):
        return self.cata(f, I)

    def left_or_map(self, f: Callable[[B], Any]):
        return self.cata(I, f)

    @property
    def ljoin(self):
        return self.right_or_map(Left)

    def __str__(self):
        return '{}({})'.format(self.__class__.__name__, self.value)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.value)

    @property
    def to_list(self):
        return self.to_maybe.to_list

    def lmap(self, f: Callable[[A], Any]):
        return Left(f(self.value)) if self.is_left else self  # type: ignore

    @property
    def get_or_raise(self):
        def fail(err):
            raise err if isinstance(err, Exception) else Exception(err)
        return self.cata(fail, I)

    def __iter__(self):
        return iter(self.to_list)

    @property
    def swap(self) -> 'Either[B, A]':
        return self.cata(Right, Left)

    @property
    def json(self):
        return self.to_maybe.json


class Right(Either):

    def __eq__(self, other):
        return isinstance(other, Right) and self.value == other.value


class Left(Either):

    def __eq__(self, other):
        return isinstance(other, Left) and self.value == other.value

__all__ = ('Either', 'Left', 'Right')



from setuptools import setup, find_packages

version_parts = (9, 3, 2)
version = '.'.join(map(str, version_parts))

gh_lenses =\
    'git+https://github.com/ingolemo/python-lenses.git#egg=lenses'

setup(
    name='amino',
    description='functional data structures and type classes',
    version=version,
    author='Torsten Schmits',
    author_email='torstenschmits@gmail.com',
    license='MIT',
    url='https://github.com/tek/amino',
    packages=find_packages(exclude=['unit', 'unit.*']),
    install_requires=[
        'fn',
        'toolz',
        'lenses',
    ],
    tests_require=[
        'spec',
        'sure',
    ],
    dependency_links=[
        gh_lenses,
    ]
)



import abc
from typing import TypeVar, Generic, Callable, Tuple

from amino.tc.base import TypeClass, tc_prop
from amino.tc.monoid import Monoid
from amino.tc.foldable import Foldable
from amino.tc.applicative import Applicative
from amino.tc.functor import Functor

F = TypeVar('F')
A = TypeVar('A')
B = TypeVar('B')


class Zip(Generic[F], TypeClass):

    @abc.abstractmethod
    def zip(self, fa: F, fb: F, *fs) -> F:
        ...

    def __and__(self, fa: F, fb: F):
        return self.zip(fa, fb)

    def apzip(self, fa: F, f: Callable[[A], B]) -> F:
        return self.zip(fa, Functor[type(fa)].map(fa, f))

    def flat_apzip(self, fa: F, f: Callable[[A], B]) -> F:
        return self.zip(fa, Functor[type(fa)].flat_map(fa, f))

    @tc_prop
    def unzip(self, fa: F) -> Tuple[F, F]:
        tpe = type(fa)
        f = Foldable[tpe]
        m = Monoid[tpe]
        a = Applicative[tpe]
        def folder(z, b):
            l, r = z
            x, y = b
            return m.combine(l, a.pure(x)), m.combine(r, a.pure(y))
        return f.fold_left(fa, (m.empty, m.empty), folder)

__all__ = ('Zip',)



import importlib
import abc
from typing import Dict, Tuple, Any, Callable, Union, Optional, GenericMeta
from functools import partial, wraps

from amino.util.string import snake_case
from amino.lazy import lazy
from amino.tc.show import Show

from amino.logging import amino_root_logger


class TypeClassMeta(GenericMeta):

    def __getitem__(self, tpe: type) -> 'TypeClass':
        return Instances.lookup(self, tpe)

    def exists_instance(self, tpe: type) -> bool:
        try:
            self[tpe]
        except ImplicitNotFound:
            return False
        else:
            return True

    exists = exists_instance


class TypeClass(metaclass=TypeClassMeta):
    pass


class TypeClasses:

    @property
    def instances(self) -> Dict[type, TypeClass]:
        pass


class ImplicitInstancesNotFound(Exception):

    def __init__(self, mod: str, cls: str, name: str) -> None:
        msg = 'invalid implicit class path {}.{} for {}'.format(mod, cls, name)
        super().__init__(msg)


operators = (
    '__floordiv__',
    '__truediv__',
    '__mod__',
    '__or__',
    '__and__',
)


class ImplicitInstances(metaclass=abc.ABCMeta):

    @lazy
    def instances(self) -> Dict[type, TypeClass]:
        return (
            TC.instances ** self._instances **  # type: ignore
            self._override_instances
        )

    @abc.abstractproperty
    def _instances(self) -> Dict[type, TypeClass]:
        ...

    @property
    def _override_instances(self) -> Dict[type, TypeClass]:
        from amino.map import Map
        return Map()


def lookup_implicit_instances(name: str, m: str, c: str) -> ImplicitInstances:
    err = ImplicitInstancesNotFound(m, c, name)
    try:
        mod = importlib.import_module(m)
    except ImportError:
        raise err
    else:
        if hasattr(mod, c):
            instances = getattr(mod, c)()
            return instances
        else:
            raise err


class InstancesMetadata:

    def __init__(self, name: str, mod: str, cls: str) -> None:
        self.name = name
        self.mod = mod
        self.cls = cls
        self._instances = None  # type: Optional[ImplicitInstances]

    def __str__(self) -> str:
        return '{}({}, {}, {})'.format(self.__class__.__name__, self.name,
                                       self.mod, self.cls)

    @property
    def instances(self) -> Dict[str, TypeClass]:
        if self._instances is None:
            self._instances = lookup_implicit_instances(self.name, self.mod,
                                                        self.cls)
        return self._instances.instances


def _infer_implicits(name: str) -> Tuple[str, str]:
    snake = snake_case(name)
    return 'amino.instances.{}'.format(snake), '{}Instances'.format(name)


class ImplicitsMeta(GenericMeta):

    @staticmethod
    def _mk_operator(name: str) -> Callable:
        def dispatch(self: Any, other: Any) -> Any:
            return self._operator(name, other)
        return dispatch

    def _attach_operators(inst) -> None:
        for op in operators:
            setattr(inst, op, ImplicitsMeta._mk_operator(op))

    def __new__(cls: type, name: str, bases: tuple, namespace: dict,
                imp_mod: str=None, imp_cls: str=None, implicits: bool=False,
                **kw: dict) -> type:
        inst = super().__new__(cls, name, bases, namespace,  # type: ignore
                               **kw)
        if not implicits:
            inst._instances_meta = None
            return inst
        else:
            if imp_mod is None or imp_cls is None:
                imp_mod, imp_cls = _infer_implicits(name)
            inst.name = name
            ImplicitsMeta._attach_operators(inst)
            meta = InstancesMetadata(name, imp_mod, imp_cls)
            inst.imp_mod = imp_mod
            inst.imp_cls = imp_cls
            inst.instances_meta = meta
            Instances.add(meta)
            return inst

    __copy__ = None


def tc_prop(f: Callable) -> Callable:
    f._tc_prop = True  # type: ignore
    return f


class Implicits(metaclass=ImplicitsMeta):
    permanent = True

    def _lookup_implicit_attr(self, name: str) -> Optional[Callable]:
        meta = type(self).instances_meta  # type: ignore
        if meta is not None:
            return next((getattr(inst, name)
                        for inst in meta.instances.v
                        if hasattr(inst, name)), None)

    def _bound_implicit_attr(self, name: str
                             ) -> Union[None, Callable, partial]:
        f = self._lookup_implicit_attr(name)
        if f is not None:
            if hasattr(f, '_tc_prop'):
                return f(self)
            else:
                return partial(f, self)
        return None

    def _set_implicit_attr(self, name: str) -> Any:
        f = self._lookup_implicit_attr(name)
        if f is not None:
            @wraps(f)
            def wrap(self: Any, *a: Any, **kw: Any) -> Any:
                return f(self, *a, **kw)  # type: ignore
            g = property(wrap) if hasattr(f, '_tc_prop') else wrap
            setattr(type(self), name, g)
            return getattr(self, name)

    def __getattr__(self, name: str) -> Callable:
        imp = (self._set_implicit_attr(name) if Implicits.permanent else
               self._bound_implicit_attr(name))
        if imp is None:
            err = '\'{}\' has no attribute \'{}\''.format(self, name)
            raise AttributeError(err)
        else:
            return imp

    def _operator(self, name: str, other: Any) -> Any:
        op = (self._set_implicit_attr(name) if Implicits.permanent else
              self._bound_implicit_attr(name))
        if op is None:
            err = '\'{}\' has no implicit operator \'{}\''.format(self, name)
            raise TypeError(err)
        else:
            return op(other)

    @property
    def dbg(self) -> Any:
        amino_root_logger.verbose(self)
        return self

    @property
    def dbgr(self) -> Any:
        v = self.log.verbose  # type: ignore
        v(repr(self))
        return self


class TypeAttrMissing(Exception):

    def __init__(self, name: str) -> None:
        msg = 'Instances class `{}` has no attribute `tpe`'.format(name)
        super().__init__(msg)


class AutoImplicitInstancesMeta(abc.ABCMeta):

    def __new__(cls: type, name: str, bases: tuple, namespace: dict) -> type:
        inst = super().__new__(cls, name, bases, namespace)
        if name != 'AutoImplicitInstances':
            imp_mod = inst.__module__
            if not hasattr(inst, 'tpe'):
                raise TypeAttrMissing(name)
            tpe_name = getattr(inst, 'tpe').__name__
            Instances.add(InstancesMetadata(tpe_name, imp_mod, name))
        return inst


class AutoImplicitInstances(ImplicitInstances,
                            metaclass=AutoImplicitInstancesMeta):
    pass


class GlobalTypeClasses(TypeClasses):

    @property
    def instances(self) -> Dict[type, TypeClass]:
        from amino.map import Map
        from amino.tc.tap import Tap
        return Map({Show: Show(), Tap: Tap()})


TC = GlobalTypeClasses()


class ImplicitNotFound(Exception):

    def __init__(self, tc: type, a: type) -> None:
        msg = 'no type class found for {}[{}]'.format(tc, a)
        super().__init__(msg)


class AllInstances:

    def __init__(self) -> None:
        self._instances = dict()  # type: Dict[str, InstancesMetadata]

    def add(self, data: InstancesMetadata) -> None:
        self._instances[data.name] = data

    def lookup(self, TC: type, G: type) -> TypeClass:
        ''' Find an instance of the type class `TC` for type `G`.
        Iterates `G`'s parent classes, looking up instances for each,
        checking whether the instance is a subclass of the target type
        class @`C`.
        '''
        from amino.lazy_list import LazyList
        from amino.anon import _, L
        match = L(self._lookup_type)(TC, _)
        result = (
            LazyList(map(match, G.__mro__))
            .find(_.is_just)
            .join
            .get_or_raise(ImplicitNotFound(TC, G))
        )
        return result[1]

    def _lookup_type(self, TC: type, G: type
                     ) -> 'amino.maybe.Maybe[TypeClass]':
        from amino.maybe import Empty
        if G.__name__ in self._instances:
            match = lambda I: isinstance(I, TC)
            return (
                self._instances[G.__name__].instances  # type: ignore
                .find(match)
            )
        else:
            return Empty()

Instances = AllInstances()  # type: AllInstances

__all__ = ('TypeClasses', 'TC', 'tc_prop', 'TypeClass')



import asyncio
from typing import Generic, TypeVar, Callable

A = TypeVar('A')
B = TypeVar('B')


class Future(Generic[A], asyncio.Future):

    def flat_map(self, f: Callable[[A], 'Future[B]']) -> 'Future[B]':
        wrapper = Future()  # type: ignore

        def cb(future: asyncio.Future):
            res = future.result()
            if res.success:
                f(res).relay_result(wrapper)
            else:
                wrapper.cancel()
        self.add_done_callback(cb)
        return wrapper

    def relay_result(self, wrapper: 'Future[A]'):
        def setter(f: asyncio.Future):
            wrapper.set_result(f.result())
        self.add_done_callback(setter)


__all__ = ('Future',)
